from .base_node import BaseNode
from .data_holder_node import DataHolderNode

import warnings


class DataNode(BaseNode):
    
    def __init__(self, node_uid, value=None, persist=False, verbose=False, alias=None):
        super(DataNode, self).__init__(node_uid, 'data', verbose, alias or 'data')
        
        self.value_holder = DataHolderNode(self.node_uid, value, self.verbose)
        self.data_persist = persist

    def set_value(self, value):
        
        self.value_holder.set_value(value)
        
    def has_value(self):
        
        return self.value_holder.has_value()
        
    def is_persisted(self):
        
        return self.data_persist

    def get(self):
        
        if self.value_holder.has_value():
            return self.value_holder.get()
        else:
            if self.verbose:
                print('computing for {}'.format(self.node_uid))
            
            self.activate_dependency_op_nodes()
            self.parent_node_weak_refs[0]().run()
            return self.value_holder.get()

    def activate_dependency_op_nodes(self):
        
        dependency_ancestor_node_weak_refs = self.get_dependency_ancestor_node_weak_refs()
        dependency_op_nodes_weak_refs = [elem for elem in dependency_ancestor_node_weak_refs 
                                  if elem().node_type == 'operation']
        
        for dependency_op_node_weak_ref in dependency_op_nodes_weak_refs:
            dependency_op_node_weak_ref().activate()

    def get_dependency_ancestor_node_weak_refs(self):

        ancestors_weak_refs = list()
        DataNode._get_dependency_ancestor_node_weak_refs(self, ancestors_weak_refs)
        return ancestors_weak_refs
    
    @staticmethod
    def _get_dependency_ancestor_node_weak_refs(self, acc):
        """
        all data nodes needed that has no values,
        all op until valued data nodes
        """

        for parent_node_weak_ref in self.get_parent_node_weak_refs():

            if parent_node_weak_ref().node_type == 'data' and parent_node_weak_ref().has_value():
                continue

            if parent_node_weak_ref().has_parent_node_weak_refs():
                
                if parent_node_weak_ref not in acc:
                    acc.append(parent_node_weak_ref)
                DataNode._get_dependency_ancestor_node_weak_refs(parent_node_weak_ref(), acc)

            else:

                acc.append(parent_node_weak_ref)

    def release_memory(self):
        
        if self.is_persisted():
            warnings.warn("You are releasing a DataNode that was persisted!", RuntimeWarning)
        
        del self.value_holder
        self.value_holder = DataHolderNode(self.node_uid, None, self.verbose)
        
        


from .base_node import BaseNode

class OperationNode(BaseNode):
    
    def __init__(self, node_uid, function, n_out, verbose=False):
        super(OperationNode, self).__init__(node_uid, 'operation', verbose, function.__name__)
        
        self.function = function
        self.n_out = n_out
        self.is_active = False
    
    def is_activated(self):
        return self.is_active
    
    def activate(self):
        
        if self.is_active:
            return
        
        if self.verbose:
            print('{} activated!'.format(self.node_uid))
            
        self.is_active = True
        
    def deactivate(self):
        
        if not self.is_active:
            return
        
        if self.verbose:
            print('{} deactivated!'.format(self.node_uid))
            
        self.is_active = False
        
    def run(self):
        
        # these strong references will be destroyed once we leave this scope
        parent_data_nodes_values = [parent_data_node_weak_ref().get() 
                                    for parent_data_node_weak_ref 
                                    in self.parent_node_weak_refs]

        if self.verbose:
            print('running {}'.format(self.node_uid))

        output_values = self.function(*parent_data_nodes_values)
        
        if self.n_out > 1:
            for i, output_value in enumerate(output_values):
                self.child_node_weak_refs[i]().set_value(output_value)
        else:
            self.child_node_weak_refs[0]().set_value(output_values)
        
        # the immediate parent data nodes
        # if any of these are needed by their child op node other than this one, 
        # we don't release them
        for parent_data_node_weak_ref in self.parent_node_weak_refs:
            
            # assume we will need to release this parent data node until proven otherwise
            release_parent_data_node = True
            
            # if the parent node is persisted, no need to check for its children op nodes
            if parent_data_node_weak_ref().is_persisted():
                continue
            
            # checking the child op nodes of the current parent data node
            for child_op_node_weak_ref in parent_data_node_weak_ref().get_child_node_weak_refs():
                
                # if one of the child op node of the parent data node is not activated, 
                # this op node gives us no reason to overturn the prejudice for releasing.
                # also, we do not care about the status of data nodes under this op node
                if not child_op_node_weak_ref().is_activated():
                    continue
                
                # now traversing down to the data nodes of the op node
                # if any of the data nodes are not filled, we need the current data node
                # to feed into that op node, so we do not want to release the memory.
                # having even one such op node + data node is enough reason to persist the 
                # parent data node (from outer most loop)
                for child_output_data_node_weak_ref in child_op_node_weak_ref().get_child_node_weak_refs():
                    
                    # if any child data node value of active op node is missing, we need the current
                    # parent data node to stay persisted
                    if not child_output_data_node_weak_ref().has_value():
                        
                        # for op node with one output data node, this weakref is referring to the
                        # output node it just computed for
                        if self.verbose:
                            print('{} still needed at {}'.format(parent_data_node_weak_ref().get_node_uid(), 
                                                                 child_op_node_weak_ref().get_node_uid()))

                        # at this point, we don't need to check any other child op node of 
                        # this parent data node.
                        # however, we may need to check the next parent_data_node_weak_ref
                        release_parent_data_node = False
                        break
                
                # if we cannot release this parent data node, no need to check other op nodes
                # just go to the next parent data node
                if not release_parent_data_node:
                    break
            
            # if the prejudice for releasing has not been overturned we will release memory
            if release_parent_data_node:
                parent_data_node_weak_ref().release_memory()
                
        self.deactivate()

